from ._version import __version__  # noqa: F401
# :obj:`str`: version

from .core import exec_sed_task, preprocess_sed_task, exec_sed_doc, exec_sedml_docs_in_combine_archive  # noqa: F401
import cbmpy


__all__ = [
    '__version__',
    'get_simulator_version',
    'exec_sed_task',
    'preprocess_sed_task',
    'exec_sed_doc',
    'exec_sedml_docs_in_combine_archive',
]


def get_simulator_version():
    """ Get the version of CBMPy

    Returns:
        :obj:`str`: version
    """
    return cbmpy.__version__



""" BioSimulators-compliant command-line interface to the `CBMPy <http://cbmpy.sourceforge.net/>`_ simulation program.

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-10-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from . import get_simulator_version
from ._version import __version__
from .core import exec_sedml_docs_in_combine_archive
from biosimulators_utils.simulator.cli import build_cli

App = build_cli('biosimulators-cbmpy', __version__,
                'CBMPy', get_simulator_version(), 'http://cbmpy.sourceforge.net/',
                exec_sedml_docs_in_combine_archive)


def main():
    with App() as app:
        app.run()



""" Utilities for working with CBMPy

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-01-02
:Copyright: 2021, BioSimulators Team
:License: MIT
"""

from .data_model import SOLVERS, OPTIMIZATION_METHODS
from biosimulators_utils.report.data_model import VariableResults
from biosimulators_utils.sedml.data_model import Variable  # noqa: F401
from biosimulators_utils.utils.core import validate_str_value, parse_value
import numpy
import types  # noqa: F401

__all__ = [
    'apply_algorithm_change_to_simulation_module_method_args',
    'apply_variables_to_simulation_module_method_args',
    'get_simulation_method_args',
    'validate_variables',
    'get_results_of_variables',
    'get_default_solver_module_function_args',
]


def apply_algorithm_change_to_simulation_module_method_args(method_props, argument_change, model, module_method_args):
    """ Set the value of an argument of a simulation method based on a SED
    algorithm parameter change

    Args:
        method_props (:obj:`dict`): properties of the simulation method
        argument_change (:obj:`AlgorithmParameterChange`): algorithm parameter change
        model (:obj:`cbmpy.CBModel.Model`): model
        module_method_args (:obj:`dict`): dictionary representing the desired simulation function,
            its parent module, and the desired keyword arguments to the function

    Raises:
        :obj:`NotImplementedError`: if the simulation method doesn't support the parameter
        :obj:`ValueError`: if the new value is not a valid value of the parameter
    """
    arg_kisao_id = argument_change.kisao_id

    parameter_props = method_props['parameters'].get(arg_kisao_id, None)
    if parameter_props is None:
        msg = "`{}` is not a parameter of {} ({}). {} suppports the following parameters:\n  - {}".format(
            arg_kisao_id, method_props['name'], method_props['kisao_id'], method_props['name'],
            '\n  - '.join('`{}`: {}'.format(param_kisao_id, method_props['parameters'][param_kisao_id]['name'])
                          for param_kisao_id in sorted(method_props['parameters'].keys()))
        )
        raise NotImplementedError(msg)

    value = argument_change.new_value
    if not validate_str_value(value, parameter_props['type']):
        msg = "`{}` is not a valid value for parameter {} ({}) of {} ({})".format(
            value, parameter_props['name'], arg_kisao_id,
            method_props['name'], method_props['kisao_id'])
        raise ValueError(msg)

    parsed_value = parse_value(value, parameter_props['type'])

    if arg_kisao_id == 'KISAO_0000553':
        solver_name = argument_change.new_value.upper()
        if solver_name not in parameter_props['enum']:
            msg = "`{}` is not a supported solver for {} ({}). The following solvers (KISAO_0000553) are available for {}:\n  - {}".format(
                argument_change.new_value, method_props['name'], method_props['kisao_id'], method_props['name'],
                '\n  - '.join('`' + solver + '`' for solver in sorted(parameter_props['enum'])))
            raise NotImplementedError(msg)

        module_method_args['solver'] = SOLVERS[solver_name]
        if not SOLVERS[solver_name]['module']:
            raise ModuleNotFoundError('{} solver ({}) is not available.'.format(argument_change.new_value, arg_kisao_id))

    elif arg_kisao_id == 'KISAO_0000552':
        if argument_change.new_value.lower() not in OPTIMIZATION_METHODS:
            msg = ("`{}` is not a supported optimization method. "
                   "The following optimization methods (KISAO_0000552) are available:\n  - {}").format(
                argument_change.new_value, '\n  - '.join('`' + name + '`' for name in sorted(OPTIMIZATION_METHODS)))
            raise NotImplementedError(msg)
        module_method_args['optimization_method'] = argument_change.new_value.lower()

    elif arg_kisao_id == 'KISAO_0000534':
        rxn_ids = set(reaction.id for reaction in model.reactions)
        desired_rxn_ids = set(parsed_value)

        invalid_rxn_ids = desired_rxn_ids.difference(rxn_ids)
        if invalid_rxn_ids:
            msg = (
                'Some of the values of {} ({}) of {} ({}) are not SBML ids of reactions:\n  - {}\n\n'
                'The values of {} should be drawn from the following list of the SMBL ids of the reactions of the model:\n  - {}'
            ).format(
                parameter_props['name'], arg_kisao_id,
                method_props['name'], method_props['kisao_id'],
                '\n  - '.join('`' + value + '`' for value in sorted(invalid_rxn_ids)),
                parameter_props['name'],
                '\n  - '.join('`' + rxn_id + '`' for rxn_id in sorted(rxn_ids)),
            )
            raise ValueError(msg)

        parsed_value = sorted(desired_rxn_ids)
        module_method_args['args'][parameter_props['arg_name']] = parsed_value

    elif arg_kisao_id == 'KISAO_0000531':
        if parsed_value < 0. or parsed_value > 1.:
            msg = 'Optimum fraction (KISAO_0000531) must be greater than or equal to 0 and less than or equal to 1, not {}.'.format(
                parsed_value)
            raise ValueError(msg)

        parsed_value *= 100.

        module_method_args['args'][parameter_props['arg_name']] = parsed_value


def apply_variables_to_simulation_module_method_args(target_sbml_id_map, method_props, variables, solver_method_args):
    """ Encode the desired output variables into arguments to simulation methods

    Args:
        target_sbml_id_map (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML id of the corresponding model object
        method_props (:obj:`dict`): properties of the simulation method
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        solver_method_args (:obj:`dict`): keyword arguments for the simulation method
    """
    if method_props['function_suffix'] == 'FluxVariabilityAnalysis':
        selected_reactions = set()
        for variable in variables:
            selected_reactions.add(target_sbml_id_map[variable.target])
        solver_method_args['selected_reactions'] = sorted(selected_reactions)


def get_simulation_method_args(method_props, module_method_args):
    """ Setup the simulation method and its keyword arguments

    Args:
        method_props (:obj:`dict`): properties of the simulation method
        module_method_args (:obj:`dict`): dictionary representing the desired simulation function,
            its parent module, and the desired keyword arguments to the function

    Returns:
        :obj:`tuple`:

            * :obj:`types.FunctionType`: simulation method
            * :obj:`dict`: keyword arguments for the simulation method
    """
    solver_props = module_method_args['solver']
    solver_module = solver_props['module']
    solver_method_name = solver_props['function_prefix'] + '_' + method_props['function_suffix']
    solver_method = getattr(solver_module, solver_method_name)
    if module_method_args['optimization_method']:
        opt_method = module_method_args['optimization_method']
        module_method_args['args']['method'] = solver_props['optimization_methods'].get(opt_method, None)
        if not module_method_args['args']['method']:
            msg = ("`{}` is not a supported optimization method of {}. "
                   "{} supports the following optimization methods (KISAO_0000552):\n  - {}").format(
                opt_method,
                solver_props['name'],
                solver_props['name'],
                '\n  - '.join('`' + name + '`' for name in sorted(solver_props['optimization_methods'].keys())))
            raise NotImplementedError(msg)

    solver_method_args = dict(**module_method_args['args'], **method_props['default_args'])

    return solver_method, solver_method_args


def validate_variables(model, method, variables, target_sbml_id_map, target_sbml_fbc_id_map, sbml_fbc_uri):
    """ Validate the desired output variables of a simulation

    Args:
        model (:obj:`cbmpy.CBModel.Model`): model
        method (:obj:`dict`): properties of desired simulation method
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        target_sbml_id_map (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML id of the corresponding model object
        target_sbml_fbc_id_map (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML-FBC id of the corresponding model object
        sbml_fbc_uri (:obj:`str`): URI for SBML FBC package
    """
    possible_target_results_path_map = set()
    for variable_pattern in method['variables']:
        for sbml_id, fbc_id, attr, _, _ in variable_pattern['get_target_results_paths'](model):
            possible_target_results_path_map.add((sbml_id, fbc_id, attr))

    invalid_symbols = set()
    invalid_targets = set()
    for variable in variables:
        if variable.symbol:
            invalid_symbols.add(variable.symbol)

        else:
            valid = True

            target = variable.target
            variable_target_id = target_sbml_id_map.get(target, None)
            variable_target_fbc_id = target_sbml_fbc_id_map.get(target, None)
            target_attr = target.partition('/@')[2] or None
            if target_attr:
                target_ns, _, target_attr = target_attr.rpartition(':')
                if target_ns and variable.target_namespaces.get(target_ns, None) != sbml_fbc_uri:
                    valid = False

            if not valid or (variable_target_id, variable_target_fbc_id, target_attr) not in possible_target_results_path_map:
                invalid_targets.add(variable.target)

    if invalid_symbols:
        msg = "{} ({}) doesn't support variables with symbols".format(
            method['name'], method['kisao_id'])
        raise NotImplementedError(msg)

    if invalid_targets:
        msg = (
            "{} ({}) doesn't support variables with the following target XPaths:\n  - {}\n\n"
            "The targets of variables should match one of the following patterns of XPaths:\n  - {}"
        ).format(
            method['name'], method['kisao_id'],
            '\n  - '.join(sorted('`' + target + '`' for target in invalid_targets)),
            '\n  - '.join(sorted('{}: `{}`'.format(
                variable_pattern['description'], variable_pattern['target'])
                for variable_pattern in method['variables']))
        )
        raise ValueError(msg)


def get_results_paths_for_variables(model, method_props, variables, target_sbml_id_map, target_sbml_fbc_id_map):
    """ Get the path to results for the desired variables

    Args:
        model (:obj:`cbmpy.CBModel.Model`): model
        method_props (:obj:`dict`): properties of desired simulation method
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        target_sbml_id_map (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML id of the corresponding model object
        target_sbml_fbc_id_map (:obj:`dict` of :obj:`str` to :obj:`str`): dictionary that maps each XPath to the
            SBML-FBC id of the corresponding model object

    Returns:
        :obj:`dict`: path to results of desired variables
    """
    possible_target_results_path_map = {}
    for variable_pattern in method_props['variables']:
        for sbml_id, fbc_id, attr, result_type, result_name in variable_pattern['get_target_results_paths'](model):
            possible_target_results_path_map[(sbml_id, fbc_id, attr)] = (result_type, result_name)

    target_results_path_map = {}
    for variable in variables:
        target = variable.target
        variable_target_id = target_sbml_id_map[target]
        variable_target_fbc_id = target_sbml_fbc_id_map[target]
        target_attr = target.partition('/@')[2].rpartition(':')[2] or None
        target_results_path_map[variable.target] = possible_target_results_path_map[(
            variable_target_id, variable_target_fbc_id, target_attr)]

    return target_results_path_map


def get_results_of_variables(target_results_path_map, method_props, solver,
                             variables, model, solution):
    """ Get the results of the desired variables

    Args:
        target_results_path_map (:obj:`dict`): path to results of desired variables
        method_props (:obj:`dict`): properties of desired simulation method
        solver (:obj:`dict`): dictionary representing the desired simulation function,
            its parent module, and the desired keyword arguments to the function
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        model (:obj:`cbmpy.CBModel.Model`): model
        solution (:obj:`object`): solution of method

    Returns:
        :obj:`VariableResults`: the results of desired variables
    """
    all_values = method_props['get_results'](method_props, solver, model, solution)

    variable_results = VariableResults()
    for variable in variables:
        result_type, result_name = target_results_path_map[variable.target]
        result = all_values[result_type].get(result_name, numpy.nan)
        variable_results[variable.id] = numpy.array(result)

    return variable_results


def get_default_solver_module_function_args(algorithm=None):
    """ Get the default solver and its default arguments for an algorithm

    Args:
        algorithm (:obj:`str`, optional): KiSAO id of algorithm

    Returns:
        :obj:`dict`: default solver and default values of its arguments
    """
    return {
        'solver': SOLVERS['GLPK'] if algorithm != 'KISAO_0000554' else SOLVERS['CPLEX'],
        'optimization_method': None,
        'args': {
            'quiet': True,
        }
    }



""" Methods for executing SED tasks in COMBINE archives and saving their outputs

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2020-10-29
:Copyright: 2020, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from .data_model import KISAO_ALGORITHMS_PARAMETERS_MAP
from .utils import (apply_algorithm_change_to_simulation_module_method_args,
                    apply_variables_to_simulation_module_method_args,
                    get_simulation_method_args, validate_variables,
                    get_results_paths_for_variables, get_results_of_variables,
                    get_default_solver_module_function_args)
from biosimulators_utils.combine.exec import exec_sedml_docs_in_archive
from biosimulators_utils.config import get_config, Config  # noqa: F401
from biosimulators_utils.log.data_model import CombineArchiveLog, TaskLog, StandardOutputErrorCapturerLevel  # noqa: F401
from biosimulators_utils.model_lang.sbml.utils import get_package_namespace as get_sbml_package_namespace
from biosimulators_utils.viz.data_model import VizFormat  # noqa: F401
from biosimulators_utils.report.data_model import ReportFormat, VariableResults, SedDocumentResults  # noqa: F401
from biosimulators_utils.sedml.data_model import (Task, ModelLanguage, ModelAttributeChange, SteadyStateSimulation,  # noqa: F401
                                                  Variable)
from biosimulators_utils.sedml import validation
from biosimulators_utils.sedml.exec import exec_sed_doc as base_exec_sed_doc
from biosimulators_utils.simulator.utils import get_algorithm_substitution_policy
from biosimulators_utils.utils.core import raise_errors_warnings
from biosimulators_utils.warnings import warn, BioSimulatorsWarning
from biosimulators_utils.xml.utils import get_namespaces_for_xml_doc
from kisao.data_model import AlgorithmSubstitutionPolicy, ALGORITHM_SUBSTITUTION_POLICY_LEVELS
from kisao.utils import get_preferred_substitute_algorithm_by_ids
from lxml import etree
import cbmpy
import copy

__all__ = [
    'exec_sedml_docs_in_combine_archive',
    'exec_sed_doc',
    'exec_sed_task',
    'preprocess_sed_task',
]


def exec_sedml_docs_in_combine_archive(archive_filename, out_dir, config=None):
    """ Execute the SED tasks defined in a COMBINE/OMEX archive and save the outputs

    Args:
        archive_filename (:obj:`str`): path to COMBINE/OMEX archive
        out_dir (:obj:`str`): path to store the outputs of the archive

            * CSV: directory in which to save outputs to files
              ``{ out_dir }/{ relative-path-to-SED-ML-file-within-archive }/{ report.id }.csv``
            * HDF5: directory in which to save a single HDF5 file (``{ out_dir }/reports.h5``),
              with reports at keys ``{ relative-path-to-SED-ML-file-within-archive }/{ report.id }`` within the HDF5 file

        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`tuple`:

            * :obj:`SedDocumentResults`: results
            * :obj:`CombineArchiveLog`: log
    """
    return exec_sedml_docs_in_archive(exec_sed_doc, archive_filename, out_dir,
                                      apply_xml_model_changes=True,
                                      config=config)


def exec_sed_doc(doc, working_dir, base_out_path, rel_out_path=None,
                 apply_xml_model_changes=True,
                 log=None, indent=0, pretty_print_modified_xml_models=False,
                 log_level=StandardOutputErrorCapturerLevel.c, config=None):
    """ Execute the tasks specified in a SED document and generate the specified outputs

    Args:
        doc (:obj:`SedDocument` or :obj:`str`): SED document or a path to SED-ML file which defines a SED document
        working_dir (:obj:`str`): working directory of the SED document (path relative to which models are located)

        base_out_path (:obj:`str`): path to store the outputs

            * CSV: directory in which to save outputs to files
              ``{base_out_path}/{rel_out_path}/{report.id}.csv``
            * HDF5: directory in which to save a single HDF5 file (``{base_out_path}/reports.h5``),
              with reports at keys ``{rel_out_path}/{report.id}`` within the HDF5 file

        rel_out_path (:obj:`str`, optional): path relative to :obj:`base_out_path` to store the outputs
        apply_xml_model_changes (:obj:`bool`, optional): if :obj:`True`, apply any model changes specified in the SED-ML file before
            calling :obj:`task_executer`.
        log (:obj:`SedDocumentLog`, optional): log of the document
        indent (:obj:`int`, optional): degree to indent status messages
        pretty_print_modified_xml_models (:obj:`bool`, optional): if :obj:`True`, pretty print modified XML models
        log_level (:obj:`StandardOutputErrorCapturerLevel`, optional): level at which to log output
        config (:obj:`Config`, optional): BioSimulators common configuration
        simulator_config (:obj:`SimulatorConfig`, optional): tellurium configuration

    Returns:
        :obj:`tuple`:

            * :obj:`ReportResults`: results of each report
            * :obj:`SedDocumentLog`: log of the document
    """
    return base_exec_sed_doc(exec_sed_task, doc, working_dir, base_out_path,
                             rel_out_path=rel_out_path,
                             apply_xml_model_changes=apply_xml_model_changes,
                             log=log,
                             indent=indent,
                             pretty_print_modified_xml_models=pretty_print_modified_xml_models,
                             log_level=log_level,
                             config=config)


def exec_sed_task(task, variables, preprocessed_task=None, log=None, config=None):
    ''' Execute a task and save its results

    Args:
        task (:obj:`Task`): task
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        preprocessed_task (:obj:`dict`, optional): preprocessed information about the task, including possible
            model changes and variables. This can be used to avoid repeatedly executing the same initialization
            for repeated calls to this method.
        log (:obj:`TaskLog`, optional): log for the task
        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`tuple`:

            :obj:`VariableResults`: results of variables
            :obj:`TaskLog`: log

    Raises:
        :obj:`ValueError`: if the task or an aspect of the task is not valid, or the requested output variables
            could not be recorded
        :obj:`NotImplementedError`: if the task is not of a supported type or involves an unsuported feature
    '''
    config = config or get_config()

    if config.LOG and not log:
        log = TaskLog()

    if preprocessed_task is None:
        preprocessed_task = preprocess_sed_task(task, variables, config=config)

    model = task.model

    # Read the model
    cbmpy_model = preprocessed_task['model']['model']

    # modify model
    if model.changes:
        raise_errors_warnings(validation.validate_model_change_types(model.changes, (ModelAttributeChange,)),
                              error_summary='Changes for model `{}` are not supported.'.format(model.id))
        model_change_setter_map = preprocessed_task['model']['model_change_setter_map']
        for change in model.changes:
            new_value = float(change.new_value)
            model_change_setter_map[change.target](new_value)

    # Set up simulation function and its keyword arguments
    variable_target_sbml_id_map = preprocessed_task['model']['variable_target_sbml_id_map']
    method_props = preprocessed_task['simulation']['method_props']
    simulation_method_args = copy.copy(preprocessed_task['simulation']['method_args'])
    apply_variables_to_simulation_module_method_args(variable_target_sbml_id_map, method_props, variables, simulation_method_args)

    # Simulate the model
    simulation_method = preprocessed_task['simulation']['method']
    solution = simulation_method(cbmpy_model, **simulation_method_args)

    # throw error if status isn't optimal
    module_method_args = preprocessed_task['simulation']['module_method_args']
    method_props['raise_if_simulation_error'](module_method_args, solution)

    # get the result of each variable
    variable_results = get_results_of_variables(
        preprocessed_task['model']['variable_target_results_path_map'],
        method_props, module_method_args['solver'],
        variables, cbmpy_model, solution)

    # log action
    if config.LOG:
        log.algorithm = preprocessed_task['simulation']['algorithm_kisao_id']
        log.simulator_details = {
            'method': simulation_method.__module__ + '.' + simulation_method.__name__,
            'arguments': simulation_method_args,
        }

    # return the result of each variable and log
    return variable_results, log


def preprocess_sed_task(task, variables, config=None):
    """ Preprocess a SED task, including its possible model changes and variables. This is useful for avoiding
    repeatedly initializing tasks on repeated calls of :obj:`exec_sed_task`.

    Args:
        task (:obj:`Task`): task
        variables (:obj:`list` of :obj:`Variable`): variables that should be recorded
        config (:obj:`Config`, optional): BioSimulators common configuration

    Returns:
        :obj:`dict`: preprocessed information about the task
    """
    config = config or get_config()

    model = task.model
    sim = task.simulation

    if config.VALIDATE_SEDML:
        raise_errors_warnings(validation.validate_task(task),
                              error_summary='Task `{}` is invalid.'.format(task.id))
        raise_errors_warnings(validation.validate_model_language(model.language, ModelLanguage.SBML),
                              error_summary='Language for model `{}` is not supported.'.format(model.id))
        raise_errors_warnings(validation.validate_model_change_types(model.changes, (ModelAttributeChange,)),
                              error_summary='Changes for model `{}` are not supported.'.format(model.id))
        raise_errors_warnings(*validation.validate_model_changes(model),
                              error_summary='Changes for model `{}` are invalid.'.format(model.id))
        raise_errors_warnings(validation.validate_simulation_type(sim, (SteadyStateSimulation, )),
                              error_summary='{} `{}` is not supported.'.format(sim.__class__.__name__, sim.id))
        raise_errors_warnings(*validation.validate_simulation(sim),
                              error_summary='Simulation `{}` is invalid.'.format(sim.id))
        raise_errors_warnings(*validation.validate_data_generator_variables(variables),
                              error_summary='Data generator variables for task `{}` are invalid.'.format(task.id))

    model_etree = etree.parse(model.source)
    model_change_sbml_id_map = validation.validate_target_xpaths(
        model.changes, model_etree, attr='id')
    variable_target_sbml_id_map = validation.validate_target_xpaths(
        variables, model_etree, attr='id')
    namespaces = get_namespaces_for_xml_doc(model_etree)
    sbml_fbc_prefix, sbml_fbc_uri = get_sbml_package_namespace('fbc', namespaces)
    variable_target_sbml_fbc_id_map = validation.validate_target_xpaths(
        variables,
        model_etree,
        attr={
            'namespace': {
                'prefix': sbml_fbc_prefix,
                'uri': sbml_fbc_uri,
            },
            'name': 'id',
        }
    )

    # Read the model
    cbmpy_model = cbmpy.CBRead.readSBML3FBC(model.source)

    # preprocess model changes
    model_change_setter_map = {}
    sbml_id_reaction_map = {rxn.id: rxn for rxn in cbmpy_model.reactions}
    invalid_changes = []
    for change in model.changes:
        target = change.target
        sbml_id = model_change_sbml_id_map[target]
        reaction = sbml_id_reaction_map.get(sbml_id, None)
        setter = None
        if reaction:
            attr = target.partition('/@')[2]
            ns, _, attr = attr.partition(':')
            ns = change.target_namespaces.get(ns, None)

            if ns == sbml_fbc_uri:
                if attr == 'lowerFluxBound':
                    setter = reaction.setLowerBound
                elif attr == 'upperFluxBound':
                    setter = reaction.setUpperBound
        if setter:
            model_change_setter_map[change.target] = setter
        else:
            invalid_changes.append(change.target)

    if invalid_changes:
        valid_changes = []
        for reaction in cbmpy_model.reactions:
            valid_changes.append(
                "/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/@fbc:lowerFluxBound".format(reaction.id))
            valid_changes.append(
                "/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction[@id='{}']/@fbc:upperFluxBound".format(reaction.id))

        msg = 'The following changes are invalid:\n  {}\n\nThe following targets are valid:\n  {}'.format(
            '\n  '.join(sorted(invalid_changes)),
            '\n  '.join(sorted(valid_changes)),
        )
        raise ValueError(msg)

    # Set up the algorithm specified by :obj:`task.simulation.algorithm.kisao_id`
    alg_substitution_policy = get_algorithm_substitution_policy(config=config)
    exec_kisao_id = get_preferred_substitute_algorithm_by_ids(
        sim.algorithm.kisao_id, KISAO_ALGORITHMS_PARAMETERS_MAP.keys(),
        substitution_policy=alg_substitution_policy)
    method_props = KISAO_ALGORITHMS_PARAMETERS_MAP[exec_kisao_id]

    # Set up the the parameters of the algorithm
    module_method_args = get_default_solver_module_function_args(exec_kisao_id)
    if exec_kisao_id == sim.algorithm.kisao_id:
        for change in sim.algorithm.changes:
            try:
                apply_algorithm_change_to_simulation_module_method_args(method_props, change, cbmpy_model, module_method_args)
            except NotImplementedError as exception:
                if (
                    ALGORITHM_SUBSTITUTION_POLICY_LEVELS[alg_substitution_policy]
                    > ALGORITHM_SUBSTITUTION_POLICY_LEVELS[AlgorithmSubstitutionPolicy.NONE]
                ):
                    warn('Unsuported algorithm parameter `{}` was ignored:\n  {}'.format(
                        change.kisao_id, str(exception).replace('\n', '\n  ')),
                        BioSimulatorsWarning)
                else:
                    raise
            except ValueError as exception:
                if (
                    ALGORITHM_SUBSTITUTION_POLICY_LEVELS[alg_substitution_policy]
                    > ALGORITHM_SUBSTITUTION_POLICY_LEVELS[AlgorithmSubstitutionPolicy.NONE]
                ):
                    warn('Unsuported value `{}` for algorithm parameter `{}` was ignored:\n  {}'.format(
                        change.new_value, change.kisao_id, str(exception).replace('\n', '\n  ')),
                        BioSimulatorsWarning)
                else:
                    raise

    # validate selected solver is available
    if not module_method_args['solver']['module']:
        raise ModuleNotFoundError('{} solver is not available.'.format(module_method_args['solver']['name']))

    # Setup simulation function and its keyword arguments
    simulation_method, simulation_method_args = get_simulation_method_args(method_props, module_method_args)

    # validate and preprocess variables
    validate_variables(cbmpy_model, method_props, variables, variable_target_sbml_id_map, variable_target_sbml_fbc_id_map, sbml_fbc_uri)
    variable_target_results_path_map = get_results_paths_for_variables(
        cbmpy_model, method_props, variables, variable_target_sbml_id_map, variable_target_sbml_fbc_id_map)

    # return preprocessed information
    return {
        'model': {
            'model': cbmpy_model,
            'model_change_setter_map': model_change_setter_map,
            'variable_target_sbml_id_map': variable_target_sbml_id_map,
            'variable_target_results_path_map': variable_target_results_path_map,
        },
        'simulation': {
            'algorithm_kisao_id': exec_kisao_id,
            'method': simulation_method,
            'method_props': method_props,
            'method_args': simulation_method_args,
            'module_method_args': module_method_args,
        }
    }



""" Data model for mapping from KiSAO terms for algorithms and their parameters
to solvers methods and their arguments.

:Author: Jonathan Karr <karr@mssm.edu>
:Date: 2021-01-02
:Copyright: 2021, Center for Reproducible Biomedical Modeling
:License: MIT
"""

from biosimulators_utils.data_model import ValueType
import cbmpy  # noqa: F401
import collections
import numpy
try:
    from cbmpy import CBGLPK
except ModuleNotFoundError:  # pragma: no cover  # GLPK should always be installed
    CBGLPK = None
try:
    from cbmpy import CBCPLEX
except ModuleNotFoundError:
    CBCPLEX = None

__all__ = [
    'SOLVERS',
    'OPTIMIZATION_METHODS',
    'FBA_OUTPUT_VARIABLE_TARGETS',
    'FVA_OUTPUT_VARIABLE_TARGETS',
    'KISAO_ALGORITHMS_PARAMETERS_MAP',
]

SOLVERS = {
    'CPLEX': {
        'name': 'CPLEX',
        'module': CBCPLEX,
        'function_prefix': 'cplx',
        'optimization_methods': {
            'auto': 'o',
            'primal': 'p',
            'dual': 'd',
            'barrier without crossover': 'b',
            'barrier': 'h',
            'sifting': 's',
            'concurrent': 'c',
        },
    },
    'GLPK': {
        'name': 'GLPK',
        'module': CBGLPK,
        'function_prefix': 'glpk',
        'optimization_methods': {
            'simplex': 's',
            'interior': 'i',
            'exact': 'e',
        },
    },
}

OPTIMIZATION_METHODS = set([
    'auto',
    'primal',
    'dual',
    'barrier without crossover',
    'barrier',
    'sifting',
    'concurrent',
    'simplex',
    'interior',
    'exact',
])

FBA_OUTPUT_VARIABLE_TARGETS = [
    {
        'description': 'objective value',
        'target_type': 'objective',
        'target': r'^/sbml:sbml/sbml:model/fbc:listOfObjectives/fbc:objective(\[.*?\])?(/@value)?$',
        'get_target_results_paths': lambda model:
            (
                [
                    (None, objective.id, None, 'objective_value', objective.id)
                    for objective in model.objectives
                ] +
                [
                    (None, objective.id, 'value', 'objective_value', objective.id)
                    for objective in model.objectives
                ]
            ),
    },
    {
        'description': 'reaction flux',
        'target_type': 'reaction',
        'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?(/@flux)?$',
        'get_target_results_paths': lambda model:
            (
                [
                    (reaction.id, None, None, 'reaction_flux', reaction.id)
                    for reaction in model.reactions
                ] +
                [
                    (reaction.id, None, 'flux', 'reaction_flux', reaction.id)
                    for reaction in model.reactions
                ]
            ),
    },
    {
        'description': 'reaction reduced cost',
        'target_type': 'reaction',
        'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@reducedCost$',
        'get_target_results_paths': lambda model:
            [
                (reaction.id, None, 'reducedCost', 'reaction_reduced_cost', reaction.id)
                for reaction in model.reactions
            ],
    },
    {
        'description': 'species shadow price',
        'target_type': 'species',
        'target': r'^/sbml:sbml/sbml:model/sbml:listOfSpecies/sbml:species(\[.*?\])?(/@shadowPrice)?$',
        'get_target_results_paths': lambda model:
            (
                [
                    (specie.id, None, None, 'species_shadow_price', specie.id)
                    for specie in model.species
                ] + [
                    (specie.id, None, 'shadowPrice', 'species_shadow_price', specie.id)
                    for specie in model.species
                ]
            ),
    },
]

FVA_OUTPUT_VARIABLE_TARGETS = [
    {
        'description': 'minimum reaction flux',
        'target_type': 'reaction',
        'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@minFlux?$',
        'get_target_results_paths': lambda model:
            [
                (reaction.id, None, 'minFlux', 'reaction_min_flux', reaction.id)
                for reaction in model.reactions
            ],
    },
    {
        'description': 'maximum reaction flux',
        'target_type': 'reaction',
        'target': r'^/sbml:sbml/sbml:model/sbml:listOfReactions/sbml:reaction(\[.*?\])?/@maxFlux?$',
        'get_target_results_paths': lambda model:
            [
                (reaction.id, None, 'maxFlux', 'reaction_max_flux', reaction.id)
                for reaction in model.reactions
            ],
    },
]


def raise_if_fba_simulation_error(module_method_args, solution):
    """ Raise an error if the solution of an FBA or pFBA simulation is not optimal

    Args:
        module_method_args (:obj:`dict`): dictionary representing the desired simulation function,
            its parent module, and the desired keyword arguments to the function
        solution (:obj:`glpk.LPX` or :obj:`cplex.Cplex`): solution of method

    Raises:
        :obj:`ValueError`: error
    """
    solver_props = module_method_args['solver']
    status_method = getattr(solver_props['module'], solver_props['function_prefix'] + '_getSolutionStatus')
    status = status_method(solution)
    if status not in ['LPS_OPT', 'MILP_OPT']:
        raise ValueError("A solution could not be found. The solver status was `{}`.".format(
            status))


def get_fba_results(method_props, solver, model, solution):
    """ Get the results of an FBA simulation

    Args:
        method_props (:obj:`dict`): properties of desired simulation method
        solver (:obj:`dict`): dictionary representing the desired simulation function,
            its parent module, and the desired keyword arguments to the function
        model (:obj:`cbmpy.CBModel.Model`): model
        solution (:obj:`glpk.LPX` or :obj:`cplex.Cplex`): solution of method

    Returns:
        :obj:`dict`: results of an FBA simulation
    """
    results = {}

    results['objective_value'] = {
        model.getActiveObjective().id: model.getObjFuncValue()
    }

    results['reaction_flux'] = model.getReactionValues()

    if method_props['kisao_id'] in ['KISAO_0000437', 'KISAO_0000528']:
        results['reaction_reduced_cost'] = getattr(solver['module'], solver['function_prefix'] + '_' + 'getReducedCosts')(solution)
    else:
        results['reaction_reduced_cost'] = {reaction.id: numpy.nan for reaction in model.reactions}

    if method_props['kisao_id'] in ['KISAO_0000437', 'KISAO_0000528'] and solver['name'] == 'CPLEX':
        results['species_shadow_price'] = {}
        for key, (lb, rhs, ub) in getattr(solver['module'], solver['function_prefix'] + '_' + 'getShadowPrices')(solution).items():
            if ub > -lb:
                results['species_shadow_price'][key] = ub
            else:
                results['species_shadow_price'][key] = lb

    else:
        results['species_shadow_price'] = {species.id: numpy.nan for species in model.species}

    return results


def get_fva_results(method_props, solver, model, solution):
    """ Get the results of an FVA simulation

    Args:
        method_props (:obj:`dict`): properties of desired simulation method
        solver (:obj:`dict`): dictionary representing the desired simulation function,
            its parent module, and the desired keyword arguments to the function
        model (:obj:`cbmpy.CBModel.Model`): model
        solution (:obj:`tuple`): solution of method

            * :obj:`numpy.ndarray`: 2-D array of FVA results.

               * Rows: reactions.
               * Columns:
                    * ``Reaction`
                    * ``Reduced Costs``
                    * ``Variability Min``
                    * ``Variability Max``
                    * ``abs(Max-Min)``
                    * ``MinStatus``
                    * ``MaxStatus``

            * :obj:`list` of :obj:`str`: reaction id for each row

    Returns:
        :obj:`dict`: results of an FVA simulation
    """
    columns = ['Reaction', 'Reduced Costs', 'Variability Min', 'Variability Max', 'abs(Max-Min)', 'MinStatus', 'MaxStatus']
    values, rxn_ids = solution
    assert values.shape[1] == len(columns)

    i_rxn_min_fluxes = columns.index('Variability Min')
    i_rxn_max_fluxes = columns.index('Variability Max')

    results = {
        'reaction_min_flux': {},
        'reaction_max_flux': {},
    }

    for i_reaction, rxn_id in enumerate(rxn_ids):
        results['reaction_min_flux'][rxn_id] = values[i_reaction, i_rxn_min_fluxes]
        results['reaction_max_flux'][rxn_id] = values[i_reaction, i_rxn_max_fluxes]

    return results


KISAO_ALGORITHMS_PARAMETERS_MAP = collections.OrderedDict([
    ('KISAO_0000437', {
        'kisao_id': 'KISAO_0000437',
        'name': 'FBA: flux balance analysis',
        'function_suffix': 'analyzeModel',
        'parameters': {
            'KISAO_0000553': {
                'name': 'solver',
                'type': ValueType.string,
                'enum': ['CPLEX', 'GLPK'],
            },
            'KISAO_0000552': {
                'name': 'optimization method',
                'type': ValueType.string,
                'enum': OPTIMIZATION_METHODS,
            },
        },
        'default_args': {
            'with_reduced_costs': True,
            'return_lp_obj': True,
        },
        'variables': FBA_OUTPUT_VARIABLE_TARGETS,
        'raise_if_simulation_error': raise_if_fba_simulation_error,
        'get_results': get_fba_results,
    }),
    ('KISAO_0000528', {
        'kisao_id': 'KISAO_0000528',
        'name': 'pFBA: parsimonious flux balance analysis (minimum sum of absolute fluxes)',
        'function_suffix': 'MinimizeSumOfAbsFluxes',
        'parameters': {
            'KISAO_0000534': {
                'name': 'selected reactions',
                'arg_name': 'selected_reactions',
                'type': ValueType.list,
            },
            'KISAO_0000531': {
                'name': 'optimum percentage',
                'arg_name': 'optPercentage',
                'type': ValueType.float,
            },
            'KISAO_0000553': {
                'name': 'solver',
                'type': ValueType.string,
                'enum': ['CPLEX', 'GLPK'],
            },
            'KISAO_0000552': {
                'name': 'optimization method',
                'type': ValueType.string,
                'enum': OPTIMIZATION_METHODS,
            },
        },
        'default_args': {
            'with_reduced_costs': True,
            'return_lp_obj': True,
        },
        'variables': FBA_OUTPUT_VARIABLE_TARGETS,
        'raise_if_simulation_error': raise_if_fba_simulation_error,
        'get_results': get_fba_results,
    }),
    ('KISAO_0000554', {
        'kisao_id': 'KISAO_0000554',
        'name': 'pFBA: parsimonious flux balance analysis (minimum number of active fluxes)',
        'function_suffix': 'MinimizeNumActiveFluxes',
        'parameters': {
            'KISAO_0000534': {
                'name': 'selected reactions',
                'arg_name': 'selected_reactions',
                'type': ValueType.list,
            },
            'KISAO_0000531': {
                'name': 'optimum percentage',
                'arg_name': 'optPercentage',
                'type': ValueType.float,
            },
            'KISAO_0000553': {
                'name': 'solver',
                'type': ValueType.string,
                'enum': ['CPLEX'],
            }
        },
        'default_args': {
            'return_lp_obj': True,
        },
        'variables': FBA_OUTPUT_VARIABLE_TARGETS,
        'raise_if_simulation_error':
            lambda module_method_args, opt_solution:
                raise_if_fba_simulation_error(module_method_args, opt_solution[1]),
        'get_results':
            lambda method_props, solver, model, opt_solution:
                get_fba_results(method_props, solver, model, opt_solution[1]),
    }),
    ('KISAO_0000526', {
        'kisao_id': 'KISAO_0000526',
        'name': 'FVA: flux variability analysis',
        'function_suffix': 'FluxVariabilityAnalysis',
        'parameters': {
            'KISAO_0000531': {
                'name': 'optimum percentage',
                'arg_name': 'optPercentage',
                'type': ValueType.float,
            },
            'KISAO_0000553': {
                'name': 'solver',
                'type': ValueType.string,
                'enum': ['CPLEX', 'GLPK'],
            },
            'KISAO_0000552': {
                'name': 'optimization method',
                'type': ValueType.string,
                'enum': OPTIMIZATION_METHODS,
            },
        },
        'default_args': {
        },
        'variables': FVA_OUTPUT_VARIABLE_TARGETS,
        'raise_if_simulation_error': lambda module_method_args, solution: None,
        'get_results': get_fva_results,
    })
])
